/**
 * Copyright (C) 2022 Xilinx, Inc
 * Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

// ------ I N C L U D E   F I L E S -------------------------------------------
// Local - Include Files
#include "MAUtilitiesCore.h"
#include "MAUtilities.h"

// 3rd Party Library - Include Files
#include <boost/tokenizer.hpp>
#include <boost/format.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/property_tree/ini_parser.hpp>
#include <boost/property_tree/json_parser.hpp>

// System - Include Files
#include <regex>
#include <thread>
#include <unistd.h> // SUDO check

#define ALIGN(x, a) (((x) + ((a) -1)) & ~((a) -1))
#define PALIGN(p, a) ((char*) (ALIGN((unsigned long long) (p), (a))))
#define GET_CELL(p) (p += 4, *((const uint32_t*) (p - 4)))

// ------ C O N S T A N T   V A R I A B L E S ---------------------------------

// ------ F U N C T I O N S ---------------------------------------------------

std::string str_available_devs(bool _inUserDomain) {
  //gather available devices for user to pick from
  boost::property_tree::ptree empty_pt;
  std::stringstream           available_devs;
  available_devs << "\nDevices present:\n";
  boost::property_tree::ptree pt = XBUtilities::get_available_devices(_inUserDomain);
  for (auto& kd : pt.get_child("devices", empty_pt)) {
    boost::property_tree::ptree& dev = kd.second;
    //available_devs << boost::format("  [%s] : %s\n") % dev.get<std::string>("bdf", "?") % dev.get<std::string>("vbnv", "?");
    available_devs << boost::format("  [ %s ] : MA35D\n") % dev.get<std::string>("bdf", "?");
  }
  return available_devs.str();
}

bool XBUtilities::dev_present(std::string bdf, bool _inUserDomain) {
  //gather available devices for user to pick from
  boost::property_tree::ptree empty_pt;
  boost::property_tree::ptree pt = XBUtilities::get_available_devices(_inUserDomain);
  for (auto& kd : pt.get_child("devices", empty_pt)) {
    boost::property_tree::ptree& dev = kd.second;
    std::string                  d1  = dev.get<std::string>("bdf", "?");
    if (d1 == bdf)
      return true;
  }
  return false;
}

void XBUtilities::list_devices(std::ostream& out, bool _inUserDomain) {
  //gather available devices for user to pick from
  boost::property_tree::ptree empty_pt;
  boost::property_tree::ptree pt = XBUtilities::get_available_devices(_inUserDomain);
  out << "Devices present:" << std::endl;
  for (auto& kd : pt.get_child("devices", empty_pt)) {
    boost::property_tree::ptree& dev = kd.second;
    std::string                  d1  = dev.get<std::string>("bdf", "?");
    out << "  [ " << d1 << " ] : MA35D" << std::endl;
  }
}

void XBUtilities::check_bdf(const std::string& bdf, bool inUserDomain) {
  //Example 0000:02:00.0
  std::regex pattern("[[:xdigit:]]+:[[:xdigit:]]+:[[:xdigit:]]+\\.[[:xdigit:]]+");
  // std::smatch matches;
  if (!std::regex_match(bdf, pattern)) {
    //std::cerr << "ERROR: Invalid BDF format. Please specify valid BDF: " << bdf << std::endl << str_available_devs(inUserDomain) << std::endl;
    throw std::runtime_error("Invalid device: " + bdf + "  (Please specify valid device in \"Domain:BDF\" format.)\n" + str_available_devs(inUserDomain));
  }
}

void XBUtilities::pt_to_map(const boost::property_tree::ptree& pt_in, uint32_t depth, const std::string& path, std::unordered_map<std::string, uint32_t>& map_out) {
  if (!path.empty())
    map_out[path] = depth;

  if (pt_in.empty()) {
    return;
  }
  ++depth;
  for (auto& kd : pt_in) {
    if (kd.first.empty()) {
      return;
    }
    std::string p = path;
    if (!p.empty()) {
      p += ".";
    }
    p += kd.first;
    pt_to_map(kd.second, depth, p, map_out);
  }
}

boost::property_tree::ptree XBUtilities::get_available_devices(bool inUserDomain) {
  boost::property_tree::ptree empty_pt;
  if (!inUserDomain)
    return empty_pt;

  std::stringstream           ss1(ma35::execute(R"({"path": "device_list"})"));
  boost::property_tree::ptree ma35_pt1;
  boost::property_tree::ptree pt;
  boost::property_tree::read_json(ss1, ma35_pt1);

  if (ma35_pt1.get<std::string>("status") == "ok") {
    boost::property_tree::ptree dev_pt;
    for (auto& device : ma35_pt1.get_child("data", empty_pt)) {
      boost::property_tree::ptree tmp_pt;
      tmp_pt.put("bdf", device.second.get_value<std::string>());
      dev_pt.push_back(std::make_pair("", tmp_pt));
    }
    pt.add_child("devices", dev_pt);
  } else
    pt.add_child("devices", empty_pt);

  return pt;
}

/* This is C++20 version of the function
std::string
get_node_path(std::unordered_map<std::string, uint32_t> &node_map, std::string &name)
{
  auto keys = std::views::keys(node_map);
  std::string path;
  uint32_t depth = std::numeric_limits<uint32_t>::max();
  auto v1 = keys | std::views::filter([&](auto& s) {return s.length() >= name.length();})
                 | std::views::filter([&](auto& s) {return s.ends_with(name);});
  //key.compare(key.length() - node_name.length(), node_name.length(), node_name)
  if (std::ranges::empty(v1))
    return {};

  for(auto& key : v1) {
    if (node_map[key] < depth) {
      path = key;
      depth = node_map[key];
    }
  }
  return path;
}
*/
std::string get_node_path(std::unordered_map<std::string, uint32_t>& node_map, std::string& name) {
  std::string path;
  bool        found = false;
  uint32_t    depth = std::numeric_limits<uint32_t>::max();

  for (const auto& [key, val] : node_map) {
    if (key.length() >= name.length()) {
      if (key.compare(key.length() - name.length(), name.length(), name) == 0) {
        if (node_map[key] < depth) {
          found = false;
          path  = key;
          depth = val;
        }
      }
    }
  }
  if (!found)
    return {};

  return path;
}

void XBUtilities::get_host_info(boost::property_tree::ptree& pt) {

  boost::property_tree::ptree ma35_pt1;
  boost::property_tree::ptree empty_pt;
  std::stringstream           ss1(ma35::execute(R"({"path": "host"})"));
  boost::property_tree::read_json(ss1, ma35_pt1);
  std::unordered_map<std::string, uint32_t> node_map;
  XBUtilities::pt_to_map(ma35_pt1, 0, std::string(""), node_map);

  pt.put("status", ma35_pt1.get<std::string>("status"));
  if (ma35_pt1.get<std::string>("status") != "ok") {
    return;
  }

  if (node_map.find("data") != node_map.end()) {
    pt.add_child("os", ma35_pt1.get_child("data", empty_pt));
    if (node_map.find("data.ma35") != node_map.end()) {
      pt.add_child("ma35", ma35_pt1.get_child("data.ma35", empty_pt));
    } else {
      std::string node_name = "ma35";
      std::string node_path = get_node_path(node_map, node_name);
      if (!node_path.empty())
        pt.add_child(node_name, ma35_pt1.get_child(node_path, empty_pt));
    }
  }
  pt.add_child("devices", XBUtilities::get_available_devices(true).get_child("devices", empty_pt));
}

void XBUtilities::get_mem_info(const std::shared_ptr<MA35Device> dev, boost::property_tree::ptree& pt) {
  std::string                 jsf_cmd1 = R"({"path": "devices/)" + dev->bdf() + R"(/telemetry)";
  std::string                 jsf_cmd2 = jsf_cmd1 + R"(/device_memory"})";
  std::string                 jsf_cmd3 = jsf_cmd1 + R"(/ddr_bandwidth"})";
  std::stringstream           ss1(ma35::execute(jsf_cmd2));
  std::stringstream           ss2(ma35::execute(jsf_cmd3));
  boost::property_tree::ptree ma35_pt1, ma35_pt2;
  boost::property_tree::ptree empty_pt;
  boost::property_tree::read_json(ss1, ma35_pt1);
  boost::property_tree::read_json(ss2, ma35_pt2);
  std::string status = "Error";
  if (ma35_pt1.get<std::string>("status") == "ok") {
    pt.add_child("memories", ma35_pt1.get_child("data", empty_pt));
    status = "ok";
  }
  if (ma35_pt2.get<std::string>("status") == "ok") {
    pt.add_child("bandwidth", ma35_pt2.get_child("data", empty_pt));
    status = "ok";
  }
  pt.put("status", status);
}

void XBUtilities::get_pcie_info(const std::shared_ptr<MA35Device> dev, boost::property_tree::ptree& pt) {
  boost::property_tree::ptree empty_pt;

  std::string                 jsf_cmd1 = R"({"path": "devices/)" + dev->bdf() + R"(/pcie_info"})";
  std::stringstream           ss1(ma35::execute(jsf_cmd1));
  boost::property_tree::ptree ma35_pt1;
  boost::property_tree::read_json(ss1, ma35_pt1);
  pt.put("status", ma35_pt1.get<std::string>("status"));
  if (ma35_pt1.get<std::string>("status") != "ok") {
    pt.add_child("pcie_info", empty_pt);
    return;
  }
  pt.add_child("pcie_info", ma35_pt1.get_child("data", empty_pt));
}

void XBUtilities::get_thermal_info(const std::shared_ptr<MA35Device> dev, boost::property_tree::ptree& pt) {
  boost::property_tree::ptree empty_pt;

  std::string                 jsf_cmd1 = R"({"path": "devices/)" + dev->bdf() + R"(/telemetry)";
  std::string                 jsf_cmd2 = jsf_cmd1 + R"(/chip/pvt_temperature"})";
  std::string                 jsf_cmd3 = jsf_cmd1 + R"(/board/temperature"})";
  std::stringstream           ss1(ma35::execute(jsf_cmd2));
  std::stringstream           ss2(ma35::execute(jsf_cmd3));
  boost::property_tree::ptree ma35_pt1, ma35_pt2;
  boost::property_tree::read_json(ss1, ma35_pt1);
  boost::property_tree::read_json(ss2, ma35_pt2);
  pt.put("status", ma35_pt1.get<std::string>("status"));
  if (ma35_pt1.get<std::string>("status") != "ok") {
    pt.add_child("temp", empty_pt);
    return;
  }
  pt.add_child("temp.chip", ma35_pt1.get_child("data", empty_pt));
  if (ma35_pt2.get<std::string>("status") != "ok") {
    return;
  }
  pt.add_child("temp.board", ma35_pt2.get_child("data", empty_pt));

  //boost::property_tree::json_parser::write_json(std::cout, pt, true /*Pretty Print*/);
}

void XBUtilities::get_device_info(const std::shared_ptr<MA35Device> dev, boost::property_tree::ptree& pt) {
  boost::property_tree::ptree empty_pt;

  std::string                 jsf_cmd1 = R"({"path": "devices/)" + dev->bdf() + R"(/device_info"})";
  std::stringstream           ss1(ma35::execute(jsf_cmd1));
  boost::property_tree::ptree ma35_pt1;
  boost::property_tree::read_json(ss1, ma35_pt1);
  pt.put("status", ma35_pt1.get<std::string>("status"));
  if (ma35_pt1.get<std::string>("status") != "ok") {
    pt.add_child("device_info", empty_pt);
    return;
  }
  pt.add_child("device_info", ma35_pt1.get_child("data", empty_pt));
}

void XBUtilities::get_error_info(const std::shared_ptr<MA35Device> dev, boost::property_tree::ptree& pt) {
  boost::property_tree::ptree empty_pt;

  std::string                 jsf_cmd1 = R"({"path": "devices/)" + dev->bdf() + R"(/telemetry/error_counters"})";
  std::stringstream           ss1(ma35::execute(jsf_cmd1));
  boost::property_tree::ptree ma35_pt1;
  boost::property_tree::read_json(ss1, ma35_pt1);
  pt.put("status", ma35_pt1.get<std::string>("status"));
  if (ma35_pt1.get<std::string>("status") != "ok") {
    pt.add_child("error_cnt", empty_pt);
    return;
  }
  pt.add_child("error_cnt", ma35_pt1.get_child("data", empty_pt));
}

void XBUtilities::do_reset(const std::shared_ptr<MA35Device> dev, std::string& type, boost::property_tree::ptree& pt) {
  std::string jsf_cmd2 = R"({"path": "devices/)" + dev->bdf() + R"(/reset", )" + R"("parameters": {"type": ")" + type + R"("}})";

  std::stringstream           ss1(ma35::execute(jsf_cmd2));
  boost::property_tree::ptree ma35_pt1;
  boost::property_tree::ptree empty_pt;
  boost::property_tree::read_json(ss1, ma35_pt1);

  pt.put("status", ma35_pt1.get<std::string>("status"));
}

void XBUtilities::do_numvfs(const std::shared_ptr<MA35Device> dev, int32_t value, boost::property_tree::ptree& pt) {
  std::string jsf_cmd2 = R"({"path": "devices/)" + dev->bdf() + R"(/numvfs", )" + R"("parameters": {"action": "numvfs", "value": )" + std::to_string(value) + R"(}})";

  boost::property_tree::ptree ma35_pt1;
  std::stringstream           ss1(ma35::execute(jsf_cmd2));
  boost::property_tree::read_json(ss1, ma35_pt1);

  pt.put("status", ma35_pt1.get<std::string>("status"));
}

void XBUtilities::get_power_info(const std::shared_ptr<MA35Device> dev, boost::property_tree::ptree& pt) {
  boost::property_tree::ptree empty_pt;

  std::string                 jsf_cmd1 = R"({"path": "devices/)" + dev->bdf() + R"(/telemetry)";
  std::string                 jsf_cmd2 = jsf_cmd1 + R"(/chip/pvt_voltage"})";
  std::string                 jsf_cmd3 = jsf_cmd1 + R"(/board"})";
  std::stringstream           ss1(ma35::execute(jsf_cmd2));
  std::stringstream           ss2(ma35::execute(jsf_cmd3));
  boost::property_tree::ptree ma35_pt1, ma35_pt2;
  boost::property_tree::read_json(ss1, ma35_pt1);
  boost::property_tree::read_json(ss2, ma35_pt2);
  pt.put("status", ma35_pt1.get<std::string>("status"));
  if (ma35_pt1.get<std::string>("status") != "ok") {
    pt.add_child("power.board", empty_pt);
    pt.add_child("power.chip", empty_pt);
    return;
  }
  pt.add_child("power.chip", ma35_pt1.get_child("data", empty_pt));
  if (ma35_pt2.get<std::string>("status") != "ok") {
    pt.add_child("power.board", empty_pt);
    return;
  }
  pt.add_child("power.board", ma35_pt2.get_child("data.voltage_current", empty_pt));
  auto& power_pt = pt.get_child("power.board");
  auto& tmp_pt   = ma35_pt2.get_child("data.power", empty_pt);
  power_pt.insert(power_pt.end(), tmp_pt.begin(), tmp_pt.end());
  //boost::property_tree::json_parser::write_json(std::cout, pt, true /*Pretty Print*/);
}

void XBUtilities::collect_devices(const std::set<std::string>& _deviceBDFs, bool _inUserDomain, DeviceCollection& _deviceCollection) {
  // -- If the collection is empty then do nothing
  if (_deviceBDFs.empty())
    return;

  for (const auto& deviceBDF : _deviceBDFs) {
    XBUtilities::check_bdf(deviceBDF, _inUserDomain);
    _deviceCollection.emplace_back(std::make_shared<MA35Device>(deviceBDF));
  }
}

void XBUtilities::can_proceed_or_throw(const std::string& info, const std::string& error) {
  std::cout << info << "\n";
  if (!XBUtilities::can_proceed(getForce()))
    throw std::runtime_error("System Error: " + error);
}

void XBUtilities::sudo_or_throw(const std::string& msg) {
  if ((getuid() == 0) || (geteuid() == 0))
    return;

  std::cerr << "ERROR: " << msg << std::endl;
  throw std::runtime_error("System Error");
}

void XBUtilities::print_exception_and_throw_cancel(const std::runtime_error& e) {
  std::cerr << boost::format("ERROR: %s\n") % e.what();
  throw std::runtime_error("Operation Cancelled");
}

static const std::map<uint64_t, std::string> oemid_map = {{0x10da, "Xilinx"}, {0x02a2, "Dell"}, {0x12a1, "IBM"}, {0xb85c, "HP"}, {0x2a7c, "Super Micro"},
    {0x4a66, "Lenovo"}, {0xbd80, "Inspur"}, {0x12eb, "Amazon"}, {0x2b79, "Google"}};

std::string XBUtilities::parse_oem_id(const std::string& oemid) {
  uint64_t          oem_id_val = 0;
  std::stringstream ss;

  try {
    oem_id_val = std::stoul(oemid, nullptr, 16);
  } catch (const std::exception&) {
    //failed to parse oemid to hex value, ignore erros and print original value
  }

  auto oemstr = oemid_map.find(oem_id_val);
  return oemstr != oemid_map.end() ? oemstr->second : "N/A";
}

uint64_t XBUtilities::string_to_base_units(std::string str, const unit& conversion_unit) {
  boost::algorithm::trim(str);

  if (str.empty())
    throw std::runtime_error("Operation Cancelled");

  int factor;
  switch (conversion_unit) {
  case unit::bytes:
    factor = 1024;
    break;
  case unit::Hertz:
    factor = 1000;
    break;
  default:
    throw std::runtime_error("Operation Cancelled");
  }

  std::string units = "B";
  if (std::isalpha(str.back())) {
    units = str.back();
    str.pop_back();
  }

  uint64_t unit_value = 0;
  boost::to_upper(units);
  if (units.compare("B") == 0)
    unit_value = 1;
  else if (units.compare("K") == 0)
    unit_value = factor;
  else if (units.compare("M") == 0)
    unit_value = factor * factor;
  else if (units.compare("G") == 0)
    unit_value = factor * factor * factor;
  else
    throw std::runtime_error("Operation Cancelled");

  boost::algorithm::trim(str);
  uint64_t size = 0;
  try {
    size = std::stoll(str);
  } catch (const std::exception& e) {
    std::cerr << "  ERROR: " << e.what() << std::endl;
    throw std::runtime_error("Operation Cancelled");
  }

  size *= unit_value;
  return size;
}

void XBUtilities::get_version_info(boost::property_tree::ptree& pt) {

  boost::property_tree::ptree ma35_pt1;
  boost::property_tree::ptree empty_pt;
  std::stringstream           ss1(ma35::execute(R"({"path": "version"})"));
  boost::property_tree::read_json(ss1, ma35_pt1);
  std::unordered_map<std::string, uint32_t> node_map;
  XBUtilities::pt_to_map(ma35_pt1, 0, std::string(""), node_map);

  pt.put("status", ma35_pt1.get<std::string>("status"));
  if (ma35_pt1.get<std::string>("status") != "ok") {
    return;
  }
  if (node_map.find("data") != node_map.end()) {
    pt.add_child("ver", ma35_pt1.get_child("data", empty_pt));
  }
}

std::string XBUtilities::get_version() {
  std::stringstream           ss;
  boost::property_tree::ptree pt;
  XBUtilities::get_version_info(pt);
  ss << boost::format("%-20s : %s\n") % "MA35D Version" % pt.get<std::string>("ver.pkg_version", "N/A");
  return ss.str();

#if 0
  std::stringstream           ss;
  boost::property_tree::ptree pt_xrt;
  boost::property_tree::ptree empty_ptree;

  ss << boost::format("%-20s : %s\n") % "Version" % pt_xrt.get<std::string>("version", "N/A");
  ss << boost::format("%-20s : %s\n") % "Branch" % pt_xrt.get<std::string>("branch", "N/A");
  ss << boost::format("%-20s : %s\n") % "Hash" % pt_xrt.get<std::string>("hash", "N/A");
  ss << boost::format("%-20s : %s\n") % "Hash Date" % pt_xrt.get<std::string>("build_date", "N/A");
  const boost::property_tree::ptree& available_drivers = pt_xrt.get_child("drivers", empty_ptree);
  for (auto& drv : available_drivers) {
    const boost::property_tree::ptree& driver   = drv.second;
    std::string                        drv_name = driver.get<std::string>("name", "N/A");
    boost::algorithm::to_upper(drv_name);
    ss << boost::format("%-20s : %s, %s\n") % drv_name % driver.get<std::string>("version", "N/A") % driver.get<std::string>("hash", "N/A");
  }
  return ss.str();
#endif
}
