/**
 * Copyright (C) 2022 Xilinx, Inc
 * Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

// ------ I N C L U D E   F I L E S -------------------------------------------
// Local - Include Files
#include "ReportElectrical.h"
#include "MAUtilities.h"

// 3rd Party Library - Include Files
#include <string>
#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>

#define BYTES_TO_MEGABYTES 0x100000ll

void ReportElectrical::getPropertyTree(const std::shared_ptr<MA35Device> dev, boost::property_tree::ptree& _pt) const {
  boost::property_tree::ptree pt;

  XBUtilities::get_power_info(dev, pt);

  // There can only be 1 root node
  _pt.add_child("electrical", pt);
}

void printPower(const boost::property_tree::ptree& pt, std::ostream& _output) {
  for (const auto& kd : pt) {
    const boost::property_tree::ptree& t1 = kd.second;
    std::string                        id = t1.get<std::string>("id", "");
    std::string                        v1 = t1.get<std::string>("volt_mV", "");
    std::string                        c1 = t1.get<std::string>("current_mA", "");
    std::string                        p1 = t1.get<std::string>("power_mW", "");
    if (id.empty())
      id = kd.first;
    if (v1.empty() && c1.empty()) {
      v1 = t1.get<std::string>("mV", "");
      c1 = t1.get<std::string>("mA", "");
    }

    if (v1.empty() && c1.empty()) {
      _output << "  id: " << id << " [" << p1 << " mW]" << std::endl;
      continue;
    }

    if (v1.empty()) {
      _output << "  id: " << id << " [" << c1 << " mA]" << std::endl;
      continue;
    }

    if (c1.empty()) {
      _output << "  id: " << id << " [" << v1 << " mV]" << std::endl;
      continue;
    }
  }
}

void ReportElectrical::writeReport(const boost::property_tree::ptree& _pt, std::ostream& _output) const {
  boost::property_tree::ptree empty_ptree;

  _output << "MA35 Electrical Info:\n";
  if (_pt.get<std::string>("electrical.status") != "ok") {
    _output << "ERROR " << _pt.get<std::string>("electrical.status") << std::endl;
    return;
  }

  try {
    _output << "Device Electrical Info:\n";
    const boost::property_tree::ptree& p1 = _pt.get_child("electrical.power.chip", empty_ptree);
    printPower(p1, _output);
    _output << "Board Electrical Info:\n";
    const boost::property_tree::ptree& p2 = _pt.get_child("electrical.power.board", empty_ptree);
    printPower(p2, _output);

  } catch (const boost::property_tree::ptree_error& ex) {
    _output << "ERROR: Unexpected error" << std::endl;
  }
  _output << std::endl;
}
