// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#pragma once

#include <memory>
#include <vector>

#include <nlohmann/json.hpp>
#include <zip.h>

#define FLASH_REGION_LIST \
  OP(ZSP) \
  OP(ESECURE) \
  OP(PCIE_CTL_PATCH) \
  OP(PCIE_PHY_PATCH) \
  OP(PCIE_FW) \
  OP(SC) \
  OP(SPI)

#define FLASH_RESULT_LIST \
  OP(OK) \
  OP(USAGE) \
  OP(MISSING_COMMAND) \
  OP(BAD_COMMAND) \
  OP(INAPPLICABLE_OPTIONS) \
  OP(EXTRA_ARGS) \
  OP(MISSING_FILE) \
  OP(MISSING_DEVICE) \
  OP(BAD_DEVICE) \
  OP(BAD_ZIP) \
  OP(FILE_ERROR) \
  OP(MISSING_HEADER) \
  OP(BAD_HEADER) \
  OP(PROGRAM_FAILED) \
  OP(VERIFY_FAILED) \
  OP(BAD_SIG) \
  OP(NO_MORE)

namespace MaFlashEnums {
#define OP(X) X,
enum result_t { FLASH_RESULT_LIST RESULT_COUNT };
enum region_t { FLASH_REGION_LIST REGION_COUNT };
#undef OP

} // namespace MaFlashEnums

class DeviceInterface {
private:
  std::string m_name;

public:
  DeviceInterface(const std::string& Name) : m_name(Name) {
  }

  const std::string& Name() const { return m_name; }

  virtual MaFlashEnums::result_t Program(MaFlashEnums::region_t Region, const std::string& Contents, bool Backup) = 0;
  virtual MaFlashEnums::result_t Verify(MaFlashEnums::region_t Region, const std::string& Contents, bool Backup)  = 0;
  virtual bool                   isAmd() const                                                                    = 0;
};

using device_t = std::unique_ptr<DeviceInterface>;

class DeviceManager {
public:
  virtual std::vector<std::string> getDeviceList()                         = 0;
  virtual device_t                 getDevice(const std::string& DeviceBdf) = 0;
};

// DeviceManager& getDeviceManager();
// void            setDeviceManager(DeviceManager& Factory);

std::unique_ptr<DeviceManager> getPcieManager(); // for default implementation or explicit hardware testing

class MaFlash {
private:
  using json = nlohmann::ordered_json;

public:
  using result_t = MaFlashEnums::result_t;
  using region_t = MaFlashEnums::region_t;
  using enum result_t;
  using enum region_t;

public:
private:
  using file_t    = std::tuple<result_t, std::string, json, std::string>;
  using zipfile_t = std::tuple<result_t, std::string, zip_file_t*, zip_int64_t>;
  using devices_t = std::vector<device_t>;

public: // for testing - pragmatic choice due to restricted scope of this class (app)
  int                            m_argc;
  char**                         m_argv;
  std::unique_ptr<DeviceManager> m_managerHolder; // to own the pcieManager.
  DeviceManager&                 m_manager;
  zip_t*                         m_zip = nullptr;
  zip_uint64_t                   m_zipNum;
  zip_uint64_t                   m_zipIndex;
  devices_t                      m_devices; // devices for operation
  std::stringstream              m_notes;

private:
  file_t    FileResult(result_t Result) { return file_t{Result, {}, {}, {}}; }
  file_t    ProcessContents(const std::string& File, std::string&& Contents);
  file_t    GetFile(const std::string& File);
  result_t  DoInfo(const std::string& File);
  zipfile_t ZipFileResult(result_t Result) { return zipfile_t{Result, {}, {}, {}}; }
  zipfile_t GetNextZipFile();
  void      DumpDeviceList();
  result_t  ValidateDevices(const std::string& DeviceString);
  result_t  DoProgramVerify(bool Verify, const std::string& File, bool Parallel, bool StopOnError, bool Backup);
  bool      VerifySignature(const json& Json, const std::string& Contents);
  void      Note(const std::string& Device, const std::string& File, result_t Result);

public:
  MaFlash(int argc, char** argv, DeviceManager* DeviceManager = nullptr) :
      m_argc(argc), m_argv(argv), m_managerHolder((!DeviceManager) ? std::move(getPcieManager()) : nullptr),
      m_manager((!DeviceManager) ? *m_managerHolder : *DeviceManager) {
      }
  ~MaFlash() {
    if (m_zip) {
      zip_close(m_zip);
    }
  }

  auto& getManager() { return m_manager; }

  result_t Run();

  friend class MaFlashTest;
};
